#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2017 Ganggao Zhu- Grupo de Sistemas Inteligentes
# gzhu[at]dit.upm.es
# DIT, UPM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from __future__ import division

from scipy.stats import spearmanr
from scipy.stats import pearsonr

from sematch.semantic.similarity import YagoTypeSimilarity
from sematch.utility import FileIO

import numpy as np
from scipy.stats import t, norm
from math import atanh, pow
from numpy import tanh

from sklearn.metrics import classification_report
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import accuracy_score
from nltk import ConfusionMatrix

def generate_report(gold, predict, labels, detailed=True):
    """
    Generate the classification report
    :param gold: the gold label
    :param predict: the predict label
    :param labels: label sets
    :return: none
    """
    if detailed:
        print 'macro averge: ', precision_recall_fscore_support(gold, predict, average='macro')
        print 'micro average: ', precision_recall_fscore_support(gold, predict, average='micro')
        print 'weighted average: ', precision_recall_fscore_support(gold, predict, average='weighted')
    else:
        import warnings
        warnings.filterwarnings("ignore")
    print 'accuracy: ', accuracy_score(gold, predict)
    print classification_report(gold, predict, target_names=labels)
    print ConfusionMatrix(gold, predict)


class SteigerTest:

    """
    This class is used for statistical test for correlations and is based on the original implementation from Philipp.

    Functions for calculating the statistical significant differences between two dependent or independent correlation
    coefficients.The Fisher and Steiger method is adopted from the R package http://personality-project.org/r/html/paired.r.html
    and is described in detail in the book 'Statistical Methods for Psychology'
    The Zou method is adopted from http://seriousstats.wordpress.com/2012/02/05/comparing-correlations/
    Credit goes to the authors of above mentioned packages!
    Author: Philipp Singer (http://www.philippsinger.info/?p=347)
    https://github.com/psinger/CorrelationStats/blob/master/corrstats.py
    """
    def __init__(self, confidence=None):
        self._confidence = confidence if confidence else 0.95

    def rz_ci(self, r, n):
        zr_se = pow(1/(n - 3), .5)
        moe = norm.ppf(1 - (1 - self._confidence)/float(2)) * zr_se
        zu = atanh(r) + moe
        zl = atanh(r) - moe
        return tanh((zl, zu))

    def rho_rxy_rxz(self, rxy, rxz, ryz):
        num = (ryz-1/2.*rxy*rxz)*(1-pow(rxy,2)-pow(rxz,2)-pow(ryz,2))+pow(ryz,3)
        den = (1 - pow(rxy,2)) * (1 - pow(rxz,2))
        return num/float(den)

    def dependent_corr(self, xy, xz, yz, n, twotailed=False, method='steiger'):
        """
        Calculates the statistic significance between two dependent correlation coefficients
        @param xy: correlation coefficient between x and y
        @param xz: correlation coefficient between x and z
        @param yz: correlation coefficient between y and z
        @param n: number of elements in x, y and z
        @param twotailed: whether to calculate a one or two tailed test, only works for 'steiger' method
        @param conf_level: confidence level, only works for 'zou' method
        @param method: defines the method uses, 'steiger' or 'zou'
        @return: t and p-val
        """
        if method == 'steiger':
            d = xy - xz
            determin = 1 - xy * xy - xz * xz - yz * yz + 2 * xy * xz * yz
            av = (xy + xz)/2
            cube = (1 - yz) * (1 - yz) * (1 - yz)

            t2 = d * np.sqrt((n - 1) * (1 + yz)/(((2 * (n - 1)/(n - 3)) * determin + av * av * cube)))
            p = 1 - t.cdf(abs(t2), n - 3)

            if twotailed:
                p *= 2

            return t2, p
        elif method == 'zou':
            L1 = self.rz_ci(xy, n)[0]
            U1 = self.rz_ci(xy, n)[1]
            L2 = self.rz_ci(xz, n)[0]
            U2 = self.rz_ci(xz, n)[1]
            rho_r12_r13 = self.rho_rxy_rxz(xy, xz, yz)
            lower = xy - xz - pow((pow((xy - L1), 2) + pow((U2 - xz), 2) - 2 * rho_r12_r13 * (xy - L1) * (U2 - xz)), 0.5)
            upper = xy - xz + pow((pow((U1 - xy), 2) + pow((xz - L2), 2) - 2 * rho_r12_r13 * (U1 - xy) * (xz - L2)), 0.5)
            return lower, upper
        else:
            raise Exception('Wrong method!')

    def independent_corr(self, xy, ab, n, n2 = None, twotailed=False, method='fisher'):
        """
        Calculates the statistic significance between two independent correlation coefficients
        @param xy: correlation coefficient between x and y
        @param xz: correlation coefficient between a and b
        @param n: number of elements in xy
        @param n2: number of elements in ab (if distinct from n)
        @param twotailed: whether to calculate a one or two tailed test, only works for 'fisher' method
        @param conf_level: confidence level, only works for 'zou' method
        @param method: defines the method uses, 'fisher' or 'zou'
        @return: z and p-val
        """

        if method == 'fisher':
            xy_z = 0.5 * np.log((1 + xy)/(1 - xy))
            ab_z = 0.5 * np.log((1 + ab)/(1 - ab))
            if n2 is None:
                n2 = n

            se_diff_r = np.sqrt(1/(n - 3) + 1/(n2 - 3))
            diff = xy_z - ab_z
            z = abs(diff / se_diff_r)
            p = (1 - norm.cdf(z))
            if twotailed:
                p *= 2

            return z, p
        elif method == 'zou':
            L1 = self.rz_ci(xy, n)[0]
            U1 = self.rz_ci(xy, n)[1]
            L2 = self.rz_ci(ab, n2)[0]
            U2 = self.rz_ci(ab, n2)[1]
            lower = xy - ab - pow((pow((xy - L1), 2) + pow((U2 - ab), 2)), 0.5)
            upper = xy - ab + pow((pow((U1 - xy), 2) + pow((ab - L2), 2)), 0.5)
            return lower, upper
        else:
            raise Exception('Wrong method!')


class WordSimDataset:
    """
    This class is used to prepare and separate word similarity datasets.
    """

    def __init__(self):
        self._yago = YagoTypeSimilarity()

    def load_dataset(self, dataset_name):
        """
         This function loads the word similarity dataset

        :param dataset_name: the file name of word similarity dataset
        :return: word pairs and huamn ratings
        """
        data = FileIO.read_list_file('dataset/wordsim/%s.txt' % dataset_name)
        #print "dataset ", dataset_name, " ", len(data), " word pairs"
        word_pairs = map(lambda x: (x.split()[0], x.split()[1]), data)
        human = map(float, map(lambda x: x.split()[2], data))
        return word_pairs, human

    def load_result(self, sim_name, dataset_name):
        """
        This function loads the result of a similarity metric for a specific dataset

        :param sim_name: the name similarity metric
        :param dataset_name: the name of word similarity dataset
        :return: cor relation score and rating scores generated by similarity metric
        """
        data = FileIO.read_list_file('dataset/wordsim/results/%s-%s.txt' % (dataset_name, sim_name))
        data = map(float, data)
        return data[0], data[1:]

    def save_result(self, cor, sim_values, sim_name, dataset_name):
        """
        This function save the result computed by a similarity metric
        :param cor: correlation with human rating
        :param sim_values: similarity scores for word pairs
        :param sim_name: the name of similarity metric
        :param dataset_name: the name of word similarity dataset
        :return:
        """
        data = ["%.3f" % cor]
        data += map(lambda x: "%.3f" % x, sim_values)
        FileIO.save_list_file('dataset/wordsim/results/%s-%s.txt' % (dataset_name, sim_name), data)


    def check_word_graph(self, w1, w2):
        """
         check if lcs word is used as type in DBpedia

        :param w1:
        :param w2:
        :return:
        """
        s1, s2 = self._yago.best_synset_pair(w1, w2)
        lcs = self._yago.least_common_subsumer(s1, s2)
        yago_concept = self._yago.synset2yago(lcs)
        graph_ic = self._yago._graph_ic.concept_ic(yago_concept)
        return True if graph_ic else False


    def check_word_type(self, w1, w2):
        """
        check if both words are used as type in DBpedia
        :param w1:
        :param w2:
        :return:
        """
        s1, s2 = self._yago.best_synset_pair(w1, w2)
        yago_concept_1 = self._yago.synset2yago(s1)
        yago_concept_2 = self._yago.synset2yago(s2)
        graph_ic_1 = self._yago._graph_ic.concept_ic(yago_concept_1)
        graph_ic_2 = self._yago._graph_ic.concept_ic(yago_concept_2)
        return True if graph_ic_1 and graph_ic_2 else False


    def check_word_noun(self, w1, w2):
        """
        check if both words are in WordNet Noun Taxonomy
        :param w1:
        :param w2:
        :return:
        """
        s1 = self._yago.word2synset(w1)
        s2 = self._yago.word2synset(w2)
        return True if s1 and s2 else False

    def separate_dataset(self, in_file, out_file, check_function):
        """
        This function is used to separate the original word similarity dataset.

        word similarity of noun: noun_rg.txt, noun_mc.txt, noun_ws353.txt, noun_ws353-sim.txt, noun_simlex.txt

        the lcs is in knowledge graph: graph_rg.txt, graph_mc.txt, graph_ws353.txt,
        graph_ws353-sim.txt, graph_simlex.txt

        both words are in knowledge graph:  type_rg.txt, type_mc.txt, type_ws353.txt, type_ws353-sim.txt,
        type_simlex.txt

        :param in_file: source dataset file
        :param out_file: target dataset file
        :param check_function: the function of mapping criteria for deciding the word pairs.
        :return:
        """
        out_data = []
        word_pairs, human = self.load_dataset(in_file)
        for i, pairs in enumerate(word_pairs):
            w1, w2 = pairs
            h = human[i]
            if check_function(w1, w2):
                out_data.append(' '.join([w1, w2, str(h)]))
        FileIO.save_list_file('dataset/wordsim/%s.txt' % out_file, out_data)

# data = FileIO.read_list_file('eval/lexical/SimLex-999.txt')
# data = data[1:]
# data_p = []
# for d in data:
#     l = d.split()
#     if l[2] == 'N':
#         data_p.append(' '.join([l[0],l[1],l[3]]))
# FileIO.save_list_file('eval/lexical/noun_simlex.txt',data_p)

class WordSimEvaluation:

    """This class is used for evaluating similarity metrics in word similarity datasets"""

    def __init__(self, metric='Spearman'):
        self._dataset = WordSimDataset()
        self._steiger_test = SteigerTest()
        self._correlation = lambda x, y: spearmanr(x, y) if metric == 'Spearman' else lambda x, y: pearsonr(x, y)

    def dataset_names(self):
        return ['noun_mc', 'noun_rg', 'noun_simlex',
                'noun_ws353', 'noun_ws353-sim', 'rg65_spanish','rg65_EN-ES']

    def evaluate_metric(self, sim_name, sim_func, dataset_name, save_results=False):
        """
        Evaluate a single similarity metric for a dataset

        :param sim_name: the name similarity metric
        :param sim_func: the similarity function
        :param dataset_name: the name of dataset
        :param save_results: True if need to save results
        :return: the correlation of metric with human ratings
        """
        word_pairs, human = self._dataset.load_dataset(dataset_name)
        sim_values = [sim_func(w1, w2) for w1, w2 in word_pairs]
        sim_values = map(lambda x: round(x, 3), sim_values)
        cor = self._correlation(sim_values, human)[0]
        cor = round(cor, 3)
        if save_results:
            self._dataset.save_result(cor, sim_values, sim_name, dataset_name)
        return cor

    def evaluate_multiple_metrics(self, sim_dict, dataset_name):
        """
        Evaluate multiple similarity metrics for a dataset

        :param sim_dict: dictionary of similarity metric name and similarity function
        :param dataset_name: the name of dataset
        :return: a dictionary containing similarity metric name and its correlation with human
        """
        cor_dict = {}
        for sim_name, sim_func in sim_dict.iteritems():
            cor_dict[sim_name] = self.evaluate_metric(sim_name, sim_func, dataset_name, save_results=True)
        return cor_dict

    def evaluate_wpath_k(self, dataset_name, ic='corpus'):
        """
        Evaluate wpath method with different k settings

        :param dataset_name: the name of dataset
        :return: a dictionary containing k settings and its correlation with human
        """
        cor_dic = {}
        for k in range(1, 11):
            if ic == 'corpus':
                sim_func = lambda x, y: self._dataset._yago.word_similarity_wpath(x, y, k/10.0)
            else:
                sim_func = lambda x, y: self._dataset._yago.word_similarity_wpath_graph(x, y, k / 10.0)
            cor_dic[k] = self.evaluate_metric('wpath-'+ str(k), sim_func, dataset_name)
        return cor_dic


    def statistical_test(self, sim_name, compare_name, dataset_name):
        """
        Steiger's Z test on two similarity metrics for a dataset

        :param sim_name: the name of similarity metric
        :param compare_name: the name of comparison similarity metric
        :param dataset_name: the name of dataset
        :return: score of Steiger's Z test and wheter significant or not
        """
        cor_x, scores_x = self._dataset.load_result(sim_name, dataset_name)
        cor_y, scores_y = self._dataset.load_result(compare_name, dataset_name)
        N = len(scores_x)
        cor_z = self._correlation(scores_x, scores_y)[0]
        cor_z = round(cor_z, 3)
        s = self._steiger_test.dependent_corr(cor_x, cor_y , cor_z, N, method='steiger')[1]
        #conf = 1 - self._steiger_test._confidence
        return cor_z, s


class AspectEvaluation:

    """
    This class is used for aspect category classification.
    """
    def __init__(self):
        pass

    def load_dataset(self):
        data = FileIO.read_json_file('dataset/aspect/data.txt')
        X, y = zip(*[(d['text'], d['label']) for d in data])
        return X, y

    def evaluate(self, X, y, classifier, detailed_report=True):
        print "Evaluating..."
        pred = classifier.classify(X)
        generate_report(list(y), list(pred), list(set(y)), detailed_report)





#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2017 Ganggao Zhu- Grupo de Sistemas Inteligentes
# gzhu[at]dit.upm.es
# DIT, UPM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from nltk.stem import WordNetLemmatizer
from nltk.corpus import wordnet as wn
from nltk.corpus import wordnet_ic
from nltk.corpus.reader.wordnet import information_content

from sematch.semantic.sparql import EntityFeatures, StatSPARQL
from sematch.semantic.graph import GraphIC
from sematch.utility import FileIO, memoized


import math
from collections import Counter

class ConceptSimilarity:
    """
    This class is used to compute taxonomical semantic similarity scores between
    concepts that are located in a concept taxonomy. A taxonomy object needs to be passed into
    this class in order to find the structural information of concepts such as depth, path length,
    and so on. The graph-based IC is needed for semantic similarity measures wpath, res, lin, jcn.
    """
    def __init__(self, taxonomy, ic_file):
        self._taxonomy = taxonomy
        self._concepts = taxonomy._nodes
        self._concept2node = taxonomy._node2id
        self._label2concepts = {label:self._concepts[i] for i, label in enumerate(taxonomy._labels)}
        self._graph_ic = GraphIC(ic_file)

    def hyponyms(self, concept):
        if concept in self._concept2node:
            nodes = self._taxonomy.hyponyms(self._concept2node[concept])
            return [self._concepts[n] for n in nodes]
        return []

    def hypernyms(self, concept):
        if concept in self._concept2node:
            nodes = self._taxonomy.hypernyms(self._concept2node[concept])
            return [self._concepts[n] for n in nodes]
        return []

    def shortest_path_length(self, concept1, concept2):
        n1 = self._concept2node[concept1]
        n2 = self._concept2node[concept2]
        return self._taxonomy.shortest_path_length(n1, n2)

    def depth(self, concept):
        if concept == 'root':
            return 1
        n = self._concept2node[concept]
        return self._taxonomy.depth(n)

    def least_common_subsumer(self, concept1, concept2):
        n1 = self._concept2node[concept1]
        n2 = self._concept2node[concept2]
        n = self._taxonomy.least_common_subsumer(n1, n2)
        if n > len(self._concepts):
            return 'root'
        return self._concepts[n]

    def method(self, name):
        def function(c1, c2):
            score = getattr(self, name)(c1, c2)
            return abs(score)
        return function

    def name2concept(self, name):
        return self._label2concepts[name] if name in self._label2concepts else []

    def concept_ic(self, concept):
        """
        Get the graph-based IC of a concept. the ic of virtual root is 0
        :param concept: the node id of concept
        :return: the ic value of concept
        """
        if concept == 'root':
            return 0.0
        else:
            return self._graph_ic.concept_ic(concept)

    @memoized
    def similarity(self, c1, c2, name='wpath'):
        """
        Compute semantic similarity between two concepts
        :param c1:
        :param c2:
        :param name:
        :return:
        """
        if c1 not in self._concept2node or c2 not in self._concept2node:
            return 'link error'
        return self.method(name)(c1, c2)

    def path(self, c1, c2):
        """
        Rada's shortest path based similarity metric
        :param c1:
        :param c2:
        :return: similarity score in [0,1]
        """
        return 1.0/ self.shortest_path_length(c1, c2)

    def wup(self, c1, c2):
        """
        Wu and Palm's similarity metric
        :param c1:
        :param c2:
        :return:
        """
        lcs = self.least_common_subsumer(c1, c2)
        depth_c1 = self.depth(c1)
        depth_c2 = self.depth(c2)
        depth_lcs = self.depth(lcs)
        return 2.0*depth_lcs / (depth_c1 + depth_c2)

    def li(self, c1, c2, alpha=0.2, beta=0.6):
        path = self.shortest_path_length(c1, c2) - 1
        lcs = self.least_common_subsumer(c1, c2)
        depth = self.depth(lcs)
        # print path, lcs, depth
        x = math.exp(-alpha * path)
        y = math.exp(beta * depth)
        # print y
        z = math.exp(-beta * depth)
        a = y - z
        b = y + z
        return x * (a / b)

    def res(self, c1, c2):
        lcs = self.least_common_subsumer(c1, c2)
        return self.concept_ic(lcs)

    def lin(self, c1, c2):
        lcs = self.least_common_subsumer(c1, c2)
        lcs_ic = self.concept_ic(lcs)
        c1_ic = self.concept_ic(c1)
        c2_ic = self.concept_ic(c2)
        combine = c1_ic + c2_ic
        if c1_ic == 0.0 or c2_ic == 0.0:
            return 0.0
        return 2.0 * lcs_ic / combine

    def jcn(self, c1, c2):
        lcs = self.least_common_subsumer(c1, c2)
        lcs_ic = self.concept_ic(lcs)
        c1_ic = self.concept_ic(c1)
        c2_ic = self.concept_ic(c2)
        lcs_ic = 2.0 * lcs_ic
        if c1_ic == 0.0 or c2_ic == 0.0:
            return 0.0
        return 1.0 / 1 + (c1_ic + c2_ic - lcs_ic)

    def wpath(self, c1, c2, k=0.8):
        lcs = self.least_common_subsumer(c1, c2)
        path = self.shortest_path_length(c1, c2) - 1
        weight = k ** self.concept_ic(lcs)
        return 1.0 / (1 + path * weight)



class WordNetSimilarity:

    """Extend the NLTK's WordNet with more similarity metrics, word lemmatization, and multilingual."""

    def __init__(self, ic_corpus='brown'):
        self._ic_corpus = wordnet_ic.ic('ic-brown.dat') if ic_corpus == 'brown' else wordnet_ic.ic('ic-semcor.dat')
        self._wn_max_depth = 19
        self._default_metrics = ['path','lch','wup','li','res','lin','jcn','wpath']
        self._wn_lemma = WordNetLemmatizer()

    def method(self, name):
        def function(syn1, syn2):
            score = getattr(self, name)(syn1, syn2)
            return abs(score)
        return function

    def synset_expand(self, s):
        result = [s]
        hypos = s.hyponyms()
        if not hypos:
            return result
        for h in hypos:
            result.extend(self.synset_expand(h))
        return result

    #return all the noun synsets in wordnet
    def get_all_synsets(self):
        return wn.all_synsets('n')

    def get_all_lemma_names(self):
        return wn.all_lemma_names('n')

    def offset2synset(self, offset):
        '''
        offset2synset('06268567-n')
        Synset('live.v.02')
        '''
        return wn._synset_from_pos_and_offset(str(offset[-1:]), int(offset[:8]))

    def synset2offset(self, ss):
        return '%08d-%s' % (ss.offset(), ss.pos())

    #semcor live%2:43:06::
    def semcor2synset(self, sense):
        return wn.lemma_from_key(sense).synset()

    def semcor2offset(self, sense):
        '''
        semcor2synset('editorial%1:10:00::')
        06268567-n
        '''
        return self.synset2offset(self.semcor2synset(sense))

    def word2synset(self, word, pos=wn.NOUN):
        word = self._wn_lemma.lemmatize(word)
        return wn.synsets(word, pos)

    def multilingual2synset(self, word, lang='spa'):
        """
        Map words in different language to wordnet synsets
        ['als', 'arb', 'cat', 'cmn', 'dan', 'eng', 'eus', 'fas', 'fin', 'fra', 'fre',
         'glg', 'heb', 'ind', 'ita', 'jpn', 'nno','nob', 'pol', 'por', 'spa', 'tha', 'zsm']
        :param word: a word in different language that has been defined in
         Open Multilingual WordNet, using ISO-639 language codes.
        :param lang: the language code defined
        :return: wordnet synsets.
        """
        return wn.synsets(word.decode('utf-8'), lang=lang, pos=wn.NOUN)


    @memoized
    def similarity(self, c1, c2, name='wpath'):
        """
        Compute semantic similarity between two concepts
        :param c1:
        :param c2:
        :param name:
        :return:
        """
        return self.method(name)(c1, c2)

    def max_synset_similarity(self, syns1, syns2, sim_metric):
        """
        Compute the maximum similarity score between two list of synsets
        :param syns1: synset list
        :param syns2: synset list
        :param sim_metric: similarity function
        :return: maximum semantic similarity score
        """
        return max([sim_metric(c1, c2) for c1 in syns1 for c2 in syns2] + [0])

    @memoized
    def word_similarity(self, w1, w2, name='wpath'):
        s1 = self.word2synset(w1)
        s2 = self.word2synset(w2)
        sim_metric = lambda x, y: self.similarity(x, y, name)
        return self.max_synset_similarity(s1, s2, sim_metric)

    @memoized
    def best_synset_pair(self, w1, w2, name='wpath'):
        s1 = self.word2synset(w1)
        s2 = self.word2synset(w2)
        sims = Counter({(c1, c2):self.similarity(c1, c2, name) for c1 in s1 for c2 in s2})
        return sims.most_common(1)[0][0]

    def word_similarity_all_metrics(self, w1, w2):
        return {m:self.word_similarity(w1, w2, name=m) for m in self._default_metrics}

    @memoized
    def word_similarity_wpath(self, w1, w2, k):
        s1 = self.word2synset(w1)
        s2 = self.word2synset(w2)
        sim_metric = lambda x, y: self.wpath(x, y, k)
        return self.max_synset_similarity(s1, s2, sim_metric)

    @memoized
    def monol_word_similarity(self, w1, w2, lang='spa', name='wpath'):
        """
         Compute mono-lingual word similarity, two words are in same language.
        :param w1: word
        :param w2: word
        :param lang: language code
        :param name: name of similarity metric
        :return: semantic similarity score
        """
        s1 = self.multilingual2synset(w1, lang)
        s2 = self.multilingual2synset(w2, lang)
        sim_metric = lambda x, y: self.similarity(x, y, name)
        return self.max_synset_similarity(s1, s2, sim_metric)

    @memoized
    def crossl_word_similarity(self, w1, w2, lang1='spa', lang2='eng', name='wpath'):
        """
         Compute cross-lingual word similarity, two words are in different language.
        :param w1: word
        :param w2: word
        :param lang1: language code for word1
        :param lang2: language code for word2
        :param name: name of similarity metric
        :return: semantic similarity score
        """
        s1 = self.multilingual2synset(w1, lang1)
        s2 = self.multilingual2synset(w2, lang2)
        sim_metric = lambda x, y: self.similarity(x, y, name)
        return self.max_synset_similarity(s1, s2, sim_metric)

    def least_common_subsumer(self, c1, c2):
        return c1.lowest_common_hypernyms(c2)[0]

    def synset_ic(self, c):
        return information_content(c, self._ic_corpus)

    def dpath(self, c1, c2, alpha=1.0, beta=1.0):
        lcs = self.least_common_subsumer(c1, c2)
        path = c1.shortest_path_distance(c2)
        path = 1.0 / (1 + path)
        path = path**alpha
        depth = lcs.max_depth() + 1
        depth = depth*1.0/(1 + self._wn_max_depth)
        depth = depth**beta
        return math.log(1+path*depth,2)

    def wpath(self, c1, c2, k=0.8):
        lcs = self.least_common_subsumer(c1,c2)
        path = c1.shortest_path_distance(c2)
        weight = k ** self.synset_ic(lcs)
        return 1.0 / (1 + path*weight)

    def li(self, c1, c2, alpha=0.2,beta=0.6):
        path = c1.shortest_path_distance(c2)
        lcs = self.least_common_subsumer(c1, c2)
        depth = lcs.max_depth()
        x = math.exp(-alpha*path)
        y = math.exp(beta*depth)
        z = math.exp(-beta*depth)
        a = y - z
        b = y + z
        return x * (a/b)

    def path(self, c1, c2):
        return c1.path_similarity(c2)

    def wup(self, c1, c2):
        return c1.wup_similarity(c2)

    def lch(self, c1, c2):
        return c1.lch_similarity(c2)

    def res(self, c1, c2):
        return c1.res_similarity(c2, self._ic_corpus)

    def jcn(self, c1, c2):
        lcs = self.least_common_subsumer(c1, c2)
        c1_ic = self.synset_ic(c1)
        c2_ic = self.synset_ic(c2)
        lcs_ic = self.synset_ic(lcs)
        diff = c1_ic + c2_ic - 2*lcs_ic
        return 1.0/(1 + diff)

    def lin(self, c1, c2):
        return c1.lin_similarity(c2, self._ic_corpus)



class YagoTypeSimilarity(WordNetSimilarity):

    """Extend the WordNet synset to linked data through YAGO mappings"""

    def __init__(self, graph_ic='models/yago_type_ic.txt', mappings="models/type-linkings.txt"):
        WordNetSimilarity.__init__(self)
        self._graph_ic = GraphIC(graph_ic)
        self._mappings = FileIO.read_json_file(mappings)
        self._id2mappings = {data['offset']: data for data in self._mappings}
        self._yago2id = {data['yago_dbpedia']: data['offset'] for data in self._mappings}

    def synset2id(self, synset):
        return str(synset.offset() + 100000000)

    def id2synset(self, offset):
        x = offset[1:]
        return wn._synset_from_pos_and_offset('n', int(x))

    def synset2mapping(self, synset, key):
        mapping_id = self.synset2id(synset)
        if mapping_id in self._id2mappings:
            mapping = self._id2mappings[mapping_id]
            return mapping[key] if key in mapping else None
        else:
            return None

    def synset2yago(self, synset):
        return self.synset2mapping(synset,'yago_dbpedia')

    def synset2dbpedia(self, synset):
        return self.synset2mapping(synset, 'dbpedia')

    def yago2synset(self, yago):
        if yago in self._yago2id:
            return self.id2synset(self._yago2id[yago])
        return None

    def word2dbpedia(self, word):
        return [self.synset2dbpedia(s) for s in self.word2synset(word) if self.synset2dbpedia(s)]

    def word2yago(self, word):
        return [self.synset2yago(s) for s in self.word2synset(word) if self.synset2yago(s)]

    def yago_similarity(self, yago1, yago2, name='wpath'):
        """
        Compute semantic similarity of two yago concepts by mapping concept uri to wordnet synset.
        :param yago1: yago concept uri
        :param yago2: yago concept uri
        :param name: name of semantic similarity metric
        :return: semantic similarity score if both uri can be mapped to synsets, otherwise 0.
        """
        s1 = self.yago2synset(yago1)
        s2 = self.yago2synset(yago2)
        if s1 and s2:
            return self.similarity(s1, s2, name)
        return 0.0

    def word_similarity_wpath_graph(self, w1, w2, k):
        s1 = self.word2synset(w1)
        s2 = self.word2synset(w2)
        return max([self.wpath_graph(c1, c2, k) for c1 in s1 for c2 in s2] + [0])

    def res_graph(self, c1, c2):
        lcs = self.least_common_subsumer(c1,c2)
        yago = self.synset2yago(lcs)
        return self._graph_ic.concept_ic(yago)

    def lin_graph(self, c1, c2):
        lcs = self.least_common_subsumer(c1,c2)
        yago_c1 = self.synset2yago(c1)
        yago_c2 = self.synset2yago(c2)
        yago_lcs = self.synset2yago(lcs)
        lcs_ic = self._graph_ic.concept_ic(yago_lcs)
        c1_ic = self._graph_ic.concept_ic(yago_c1)
        c2_ic = self._graph_ic.concept_ic(yago_c2)
        combine = c1_ic + c2_ic
        if c1_ic == 0.0 or c2_ic == 0.0:
            return 0.0
        return 2.0 * lcs_ic / combine

    def jcn_graph(self, c1, c2):
        lcs = self.least_common_subsumer(c1,c2)
        yago_c1 = self.synset2yago(c1)
        yago_c2 = self.synset2yago(c2)
        yago_lcs = self.synset2yago(lcs)
        lcs_ic = self._graph_ic.concept_ic(yago_lcs)
        c1_ic = self._graph_ic.concept_ic(yago_c1)
        c2_ic = self._graph_ic.concept_ic(yago_c2)
        lcs_ic = 2.0 * lcs_ic
        if c1_ic == 0.0 or c2_ic == 0.0:
            return 0.0
        return 1.0 / 1+(c1_ic + c2_ic - lcs_ic)

    def wpath_graph(self, c1, c2, k=0.9):
        lcs = self.least_common_subsumer(c1, c2)
        path = c1.shortest_path_distance(c2)
        yago_lcs = self.synset2yago(lcs)
        weight = k ** self._graph_ic.concept_ic(yago_lcs)
        return 1.0 / (1 + path*weight)


class EntitySimilarity:

    """This class implements entity relatedness using DBpedia links and entity concepts"""

    def __init__(self):
        self._features = EntityFeatures()
        self._stats = StatSPARQL()
        self._yago = YagoTypeSimilarity()

    def similarity(self, entity1, entity2):
        concepts_1 = self._features.type(entity1)
        concepts_1 = [c for c in concepts_1 if c.__contains__('class/yago')]
        concepts_2 = self._features.type(entity2)
        concepts_2 = [c for c in concepts_2 if c.__contains__('class/yago')]
        synsets_1 = [self._yago.yago2synset(c) for c in concepts_1 if self._yago.yago2synset(c)]
        synsets_2 = [self._yago.yago2synset(c) for c in concepts_2 if self._yago.yago2synset(c)]
        if not synsets_1 or not synsets_2:
            return 0.0
        s1,_ = zip(*Counter({s:self._yago.synset_ic(s) for s in synsets_1}).most_common(5))
        s2,_ = zip(*Counter({s:self._yago.synset_ic(s) for s in synsets_2}).most_common(5))
        N1 = len(s1)
        N2 = len(s2)
        score1 = sum([max([self._yago.similarity(syn1, syn2) for syn2 in s2]) for syn1 in s1]) / N1
        score2 = sum([max([self._yago.similarity(syn1, syn2) for syn1 in s1]) for syn2 in s2]) / N2
        return (score1 + score2) / 2.0

    def relatedness(self, entity1, entity2):
        ab = self._stats.entity_share(entity1, entity2)
        if ab == 0:
            return 0
        a = self._stats.entity_relation(entity1)
        b = self._stats.entity_relation(entity2)
        x = math.log(max([a,b])) - math.log(ab)
        y = math.log(self._stats.entity_N()) - math.log(min([a,b]))
        return x / y




import json
import os


class FileIO:

    @staticmethod
    def path():
        return os.path.dirname(__file__)

    @staticmethod
    def filename(name):
        if FileIO.path() not in name:
            name = os.path.join(FileIO.path(), name)
        return name

    @staticmethod
    def read_json_file(name):
        data = []
        with open(FileIO.filename(name),'r') as f:
            for line in f:
                data.append(json.loads(line))
        return data

    @staticmethod
    def save_json_file(name, data):
        with open(FileIO.filename(name), 'w') as f:
            for d in data:
                json.dump(d, f)
                f.write("\n")

    @staticmethod
    def append_json_file(name, data):
        with open(FileIO.filename(name), 'a') as f:
            for d in data:
                json.dump(d, f)
                f.write("\n")

    @staticmethod
    def append_list_file(name, data):
        with open(FileIO.filename(name), 'a') as f:
            for d in data:
                f.write(d)
                f.write('\n')

    @staticmethod
    def save_list_file(name, data):
        with open(FileIO.filename(name),'w') as f:
            for d in data:
                f.write(d)
                f.write('\n')

    @staticmethod
    def read_list_file(name):
        with open(FileIO.filename(name),'r') as f:
            data = [line.strip() for line in f]
        return data


import collections
import functools

class memoized(object):
   '''Decorator. Caches a function's return value each time it is called.
   If called later with the same arguments, the cached value is returned
   (not reevaluated).
   '''
   def __init__(self, func):
      self.func = func
      self.cache = {}
   def __call__(self, *args):
      if not isinstance(args, collections.Hashable):
         # uncacheable. a list, for instance.
         # better to not cache than blow up.
         return self.func(*args)
      if args in self.cache:
         return self.cache[args]
      else:
         value = self.func(*args)
         self.cache[args] = value
         return value
   def __repr__(self):
      '''Return the function's docstring.'''
      return self.func.__doc__
   def __get__(self, obj, objtype):
      '''Support instance methods.'''
      return functools.partial(self.__call__, obj)




#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2017 Ganggao Zhu- Grupo de Sistemas Inteligentes
# gzhu[at]dit.upm.es
# DIT, UPM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from abc import ABCMeta, abstractmethod
from collections import deque
from itertools import combinations

import networkx as nx
import numpy as np

from sematch.semantic.ontology import DBpedia
from sematch.utility import FileIO


class DataTransform:
    """
    The interface of taxonomy data transformation. From skos, rdfs or any format, to
    nodes, edges, and labels.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def transform(self):
        """return nodes, labels, and edges"""
        pass

class DBpediaDataTransform(DataTransform):

    def __init__(self):
        self._ontology = DBpedia()

    def transform(self):
        nodes =  map(lambda x:x.toPython(), self._ontology.classes)
        node_id = {n:i for i,n in enumerate(nodes)}
        labels = [self._ontology.token(value) for i,value in enumerate(self._ontology.classes)]
        edges = []
        for i, node in enumerate(nodes):
            children = self._ontology.subClass(node)
            children = [child for child in children if child in nodes]
            children_ids = map(lambda x:node_id[x], children)
            for child_id in children_ids:
                edges.append((i, child_id))
        return nodes, labels, edges

class Taxonomy:

    def __init__(self, dt):
        self._nodes, self._labels, self._edges = dt.transform()
        self._node2id = {value:i for i,value in enumerate(self._nodes)}
        self._label2id = {value:i for i, value in enumerate(self._labels)}
        #virtual root
        self._root = len(self._nodes) + 1
        self._taxonomy = nx.Graph()
        self._hyponyms = {}
        self._hypernyms = {}
        self.build_graph()

    def tree_encoder(self):
        # breadth first tree traverse to encode nodes
        # root is encoded as '#', a subnode of root is '#01' if code_length is 2
        # then a 3rd level node is '#01#01'
        # we can use string to encode tree strcuture for those trees without multiple inherences.
        id2code = {}
        id2code[self._root] = 'root'
        children_lengths = map(lambda x: len(self._hyponyms[x]), self._hyponyms.keys())
        n = max(children_lengths)
        code_length = 1
        for i in range(1, 8):
            if i * 10 - n > 0:
                break
            code_length = i
        coder = lambda x: "#%0*d" % (code_length, x)
        queue = deque([self._root])
        while queue:
            node = queue.popleft()
            if node in self._hyponyms:
                node_code = id2code[node]
                hypos = self._hyponyms[node]
                for i, child in enumerate(hypos):
                    queue.append(child)
                    id2code[child] = node_code + coder(i)
        return id2code


    def build_graph(self):
        map(self._taxonomy.add_node, range(self._root))
        parents, children = zip(*self._edges)
        parents_set = set(parents)
        children_set = set(children)
        root_children = []
        #to find out those node has no edges
        for i, n in enumerate(self._nodes):
            if i not in parents_set and i not in children_set:
                root_children.append(i)
        #find out those parents that are not appeared in children set
        for p in parents_set:
            if p not in children_set:
                root_children.append(p)
        for node in root_children:
            self._taxonomy.add_edge(self._root, node)
        for parent, child in self._edges:
            self._taxonomy.add_edge(parent, child)
        for parent, child in self._edges:
            self._hyponyms.setdefault(parent,[]).append(child)
        self._hyponyms[self._root] = root_children
        for parent, child in self._edges:
            self._hypernyms.setdefault(child, []).append(parent)
        for n in root_children:
            self._hypernyms[n] = [self._root]

    def shortest_path_length(self, node1, node2):
        return len(nx.shortest_path(self._taxonomy, node1, node2))

    def depth(self, node):
        return self.shortest_path_length(self._root, node)

    def least_common_subsumer(self, node1, node2):
        path1 = nx.shortest_path(self._taxonomy, node1, self._root)
        path2 = nx.shortest_path(self._taxonomy, node2, self._root)
        i = 1
        lcs = self._root
        while i <= len(path1) and i <= len(path2):
            if path1[-i] == path2[-i]:
                lcs = path1[-i]
            i = i + 1
        return lcs

    def hyponyms(self, node):
        return self._hyponyms[node] if node in self._hyponyms else []

    def hypernyms(self, node):
        return self._hypernyms[node] if node in self._hypernyms else []

class SimGraph:
    '''
    General Purpose Similarity Graph which is used to perform graph analysis.

    The nodes in graph represent general purpose object such as concept, word, phrase,
    sentence, document, entity.

    The edges in graph represent semantic similarity between nodes. The similarity score between
    objects is ranged in [0,1] and computed by specific similarity measures.

    '''

    def __init__(self, nodes, sim_metric, threshold=0.0):
        self._nodes = nodes
        self._sim_metric = sim_metric
        self._threshold = threshold
        self._graph = self.similarity_graph(self.similarity_matrix())

    def similarity_matrix(self):
        """
        :return: a similarity matrix
        """
        N = len(self._nodes)
        M = np.zeros((N, N), dtype=np.float64)
        for i in range(N):
            M[i,i] = 1.0
        for x,y in combinations(range(N), 2):
            score = self._sim_metric(self._nodes[x], self._nodes[y])
            if score > self._threshold:
                M[x,y] = score
            else:
                M[x,y] = 0.0
        return M

    def similarity_graph(self, M):
        return nx.from_numpy_matrix(M)

    def page_rank(self):
        rank_scores = nx.pagerank(self._graph)
        return {self._nodes[key]: value for key, value in rank_scores.iteritems()}

    def hits(self):
        '''h, a = hits() hub and authority'''
        return nx.hits(self._graph)

    def minimum_spanning_tree(self):
        return nx.minimum_spanning_tree(self._graph)

from sematch.semantic.sparql import StatSPARQL
import math

class GraphIC:

    """
    This class is used to compute graph-based IC in knowledge graph, which is
    basically the proportion of instances tagged with a specific concept
    """

    def __init__(self, ic_file):
        self._ic_file = ic_file
        self._graph_ic = self.graph_ic_reader(ic_file)
        self._graph_stats = StatSPARQL()
        self._N = self._graph_stats.entity_N()

    def concept_ic(self, concept):
        """
        Compute the ic value of a concept using sparql query
        :param concept: a id of concept, here is the uri of concept
        :return: the ic value of the concept
        """
        if concept in self._graph_ic:
            return self._graph_ic[concept]
        else:
            freq = int(self._graph_stats.concept_freq(concept))
            if freq == 0:
                ic = 0.0
            else:
                prob = 1.0 * freq / self._N
                ic = -math.log(prob)
            self.graph_ic_writer(self._ic_file, [{'concept':concept, 'ic':str(ic)}])
            self._graph_ic[concept] = ic
            return ic

    def graph_ic_reader(self, filename):
        """
        Load the saved IC values
        :param filename: the file containing IC values of concepts
        :return: a dictionary concept:IC
        """
        data = FileIO.read_json_file(filename)
        return {d['concept']:float(d['ic']) for d in data}

    def graph_ic_writer(self, filename, data):
        """
        Save the ic values for a concept for faster access.
        :param filename:
        :param data:
        :return:
        """
        FileIO.append_json_file(filename, data)





#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2017 Ganggao Zhu- Grupo de Sistemas Inteligentes
# gzhu[at]dit.upm.es
# DIT, UPM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from SPARQLWrapper import SPARQLWrapper, JSON
from rdflib import RDF, RDFS, OWL

class BaseSPARQL:

    """This class implements basic sparql patterns."""

    def __init__(self, url="http://dbpedia.org/sparql", limit=5000):
        self._url = url
        self._sparql = SPARQLWrapper(url)
        self._sparql.setReturnFormat(JSON)
        self._tpl = """SELECT DISTINCT %s WHERE {\n\t%s\n} LIMIT """ + str(limit)
        self._count_tpl = """SELECT %s WHERE {\n\t%s\n} LIMIT """ + str(limit)
        self._text_tpl = """SELECT DISTINCT %s, ?label, ?abstract WHERE {\n\t%s\n} LIMIT """ + str(limit)

    def execution(self, query, show_query=False):
        if show_query:
            print query
        self._sparql.setQuery(query)
        results = self._sparql.query().convert()
        #print results
        return results["results"]["bindings"]

    #SELECT DISTINCT ?x WHERE is query line containing variables
    #{?s ?p ?o . ?s2 ?p2 ?o } is triples line of query

    def execution_template(self, variable, query, triples, template, show_query=False):
        """retrieve query results from sparql end point"""
        return [r[variable]["value"] for r in self.execution(template % (query, triples), show_query)]

    def create_query(self, variable, triples):
        """create a sparql query in string based on template, variable and triples"""
        return self._tpl % (self.q_mark(variable), triples)

    def resource_query(self, variable, triples, show_query=False):
        """execute query to return resources"""
        return self.execution_template(variable, self.q_mark(variable), triples, self._tpl, show_query)

    def text_query(self, variable, triples, lang='en', show_query=False):
        triples += self.label_triple(variable, lang)
        triples += self.abstract_triple(variable, lang)
        data = self.execution(self._text_tpl % (self.q_mark(variable), triples), show_query)
        result = []
        for d in data:
            res = {}
            res['uri'] = d[variable]['value']
            res['label'] = d['label']['value']
            res['abstract'] = d['abstract']['value']
            result.append(res)
        return result

    def count_query(self, variable, triples):
        """execute query to return counts """
        query = 'count(%s) as %s' % (self.q_mark(variable), self.q_mark(variable))
        return self.execution_template(variable, query, triples, self._count_tpl)[0]

    def literal(self, label, lang='en'):
        return '"%s"@%s' % (label, lang)

    def triple(self, subject, predicate, object):
        return """ \n\t%s %s %s .""" % (subject, predicate, object)

    def union(self, triples):
        triples = map(lambda x: """{ %s }""" % x, triples)
        return "\n UNION ".join(triples)

    def lang_filter(self, variable, lang):
        return """ \n\tFILTER( lang(%s) = "%s") .""" % (self.q_mark(variable), lang)

    def regex_filter(self, variable, expression):
        return """\n\tFILTER regex(%s, "%s") .""" % (variable, expression)

    def lcase_filter(self, variable, expression):
        return """\n\tFILTER(lcase(str(%s)) = '%s') .""" % (variable, expression)

    def q_mark(self, x):
        return '?%s' % x

    def composeX(self, f, g):
        return lambda x: f(g(x))

    def composeXY(self, f, g):
        return lambda x, y: f(y, g(x))

    def uri(self, x):
        return '<%s>' % x

    def new_triple(self, v, triple):
        def function(triple_type, x, y):
            v2, triple2 = triple_type(x, y, v)
            return v2, triple + triple2
        return function

    def v_triple(self, s, p, o):
        """unknown s, p, o"""
        return self.triple(self.q_mark(s), self.q_mark(p), self.q_mark(o))

    def s_triple(self, s, p, o):
        """known s, unknown p, o"""
        return self.triple(self.uri(s), self.q_mark(p), self.q_mark(o))

    def p_triple(self, s, p, o):
        """known p, unknown s, o"""
        return self.triple(self.q_mark(s), self.uri(p), self.q_mark(o))

    def o_triple(self, s, p, o):
        """known o, unknown s, p"""
        return self.triple(self.q_mark(s), self.q_mark(p), self.uri(o))

    def sp_triple(self, s, p, v):
        """given known subject and predicate, object is a variable"""
        return v, self.triple(self.uri(s), self.uri(p), self.q_mark(v))

    def po_triple(self, p, o, v):
        """given known predicate and object, subject is a variable"""
        return v, self.triple(self.q_mark(v), self.uri(p), self.uri(o))

    def so_triple(self, s, o, v):
        """given known subject and object, predicate is a variable"""
        return v, self.triple(self.uri(s), self.q_mark(v), self.uri(o))

    def label_triple(self, s, lang='en'):
        return self.triple(self.q_mark(s),
                           self.uri(RDFS.label),
                           self.q_mark('label')) + self.lang_filter('label', lang)

    def abstract_triple(self, s, lang='en'):
        return self.triple(self.q_mark(s),
                              self.uri('http://dbpedia.org/ontology/abstract'),
                              self.q_mark('abstract')) + self.lang_filter('abstract', lang)

    def thing(self, s):
        return self.po_triple(RDF.type, OWL.Thing, s)

    def type_of_thing(self, concept, v):
        """given the type such as movie, return a list of movien entities"""
        return self.new_triple(*self.thing(v))(self.po_triple, RDF.type, concept)

    def domain(self, property):
        return self.resource_query(*self.sp_triple(property, RDFS.domain, 'o'))

    def range(self, property):
        return self.resource_query(*self.sp_triple(property, RDFS.range, 'o'))

    def subclass(self, concept):
        return self.resource_query(*self.po_triple(RDFS.subClassOf, concept, 's'))


class EntityFeatures(BaseSPARQL):

    """This class implements entity feature extraction from DBpedia through SPARQL,
    which supports features such as entity types, labels, abstracts, categories"""

    def __init__(self):
        BaseSPARQL.__init__(self)

    def label(self, entity, lang='en'):
        """query the lable of a resource"""
        return self.resource_query(*self.new_triple('o',
               self.lang_filter('o', lang))(self.sp_triple, entity, RDFS.label))

    def type(self, entity):
        return self.resource_query(*self.sp_triple(entity, RDF.type, 'o'))

    def category(self, entity):
        return self.resource_query(*self.sp_triple(entity,
                                'http://purl.org/dc/terms/subject', 'o'))

    def abstract(self, entity, lang='en'):
        return self.resource_query(*self.new_triple('o',
            self.lang_filter('o', lang))(self.sp_triple, entity,
                'http://dbpedia.org/ontology/abstract'))

    def features(self, entity):
        return {
            'label':self.label(entity)[0],
            'type':self.type(entity),
            'category':self.category(entity),
            'abstract':self.abstract(entity)[0]
        }

class StatSPARQL(BaseSPARQL):

    """This class implements statistics for concepts and entities using SPARQL. It can be used to
    count concept frequency, entity relation frequency from DBpedia."""

    def __init__(self):
        BaseSPARQL.__init__(self)

    def entity_N(self, cal=False, default=4298433):
        """select count(?s) where ?s is thing"""
        if cal:
            return int(self.count_query(*self.thing('s')))
        return default

    def concept_freq(self, concept):
        """select count(?s) where ?s is concept. ?s is thing"""
        return self.count_query(*self.type_of_thing(concept, 's'))

    # select count(?p) where ?s1 is c1 . ?s2 is c2 . ?s1 ?p ?s2 or ?s2 ?p ?s1
    # the types of two subjects are known
    def concept_coocurence(self, c1, c2):
        s1, t1 = self.type_of_thing(c1, 's1')
        s2, t2 = self.type_of_thing(c2, 's2')
        t3 = self.v_triple(s1, 'p', s2)
        t4 = self.v_triple(s2, 'p', s1)
        count_1 = self.count_query('p', t1 + t2 + t3)
        count_2 = self.count_query('p', t1 + t2 + t4)
        return int(count_1) + int(count_2)

    # select count(?p) where ?s is concept ?s ?p ?o or ?o ?p ?s
    def concept_relation(self, concept):
        s1, t1 = self.type_of_thing(concept, 's1')
        s2, t2 = self.thing('s2')
        t3 = self.v_triple(s1, 'p', s2)
        t4 = self.v_triple(s2, 'p', s1)
        count_1 = self.count_query('p', t1 + t2 + t3)
        count_2 = self.count_query('p', t1 + t2 + t4)
        return int(count_1) + int(count_2)

    # select count(?o) where s ?p ?o or ?o ?p s
    # subject is known
    def entity_relation(self, entity):
        s1, t1 = self.thing('s1')
        t2 = self.s_triple(entity, 'p', s1)
        t3 = self.o_triple(s1, 'p', entity)
        count_1 = self.count_query(s1, t1 + t2)
        count_2 = self.count_query(s1, t1 + t3)
        return int(count_1) + int(count_2)

    def entity_share(self, entity1, entity2):
        s1, t1 = self.thing('s1')
        t2 = self.s_triple(entity1, 'p1', s1 )
        t3 = self.o_triple(s1, 'p1', entity1)
        t4 = self.s_triple(entity2, 'p2', s1)
        t5 = self.o_triple(s1, 'p2', entity2)
        count_1 = self.count_query(s1, t1+t2+t4)
        count_2 = self.count_query(s1, t1+t2+t5)
        count_3 = self.count_query(s1, t1+t3+t4)
        count_4 = self.count_query(s1, t1+t3+t5)
        return sum(map(int, [count_1, count_2, count_3, count_4]))


class NameSPARQL(BaseSPARQL):

    """This class implements exact name matching for entities. It queries entity resource
    from DBpedia through entity's names"""

    def __init__(self):
        BaseSPARQL.__init__(self)
        self.name = 'http://dbpedia.org/property/name'
        self.commonName = 'http://dbpedia.org/property/commonName'
        self.longName = 'http://dbpedia.org/ontology/longName'
        self.demonym = 'http://dbpedia.org/ontology/demonym'
        self.acronym = 'http://dbpedia.org/property/acronym'
        self.redirect_pages = 'http://dbpedia.org/ontology/wikiPageRedirects'
        self.disambiguation_pages = 'http://dbpedia.org/ontology/wikiPageDisambiguates'

    def wiki2dbpedia(self, wiki_uri):
        if wiki_uri.__contains__('https://en.wikipedia.org/wiki/'):
            wiki_uri = wiki_uri.replace('https:', 'http:')
            q = self.triple('?s', self.uri('http://xmlns.com/foaf/0.1/isPrimaryTopicOf'), self.uri(wiki_uri))
            return self.resource_query('s', q)
        return []

    def redirect(self, name):
        t = self.triple('?x', self.uri(RDFS.label), self.literal(name))
        return t + self.triple('?x', self.uri(self.redirect_pages), '?s')

    def check_redirect(self, dbr):
        return self.resource_query(*self.sp_triple(dbr, self.redirect_pages, 's'))

    # check dbpedia resource, no redirect page,
    def resource_filter(self, data):
        result = [d for d in data if d.__contains__('http://dbpedia.org/resource/')]
        # remove categories
        result = [d for d in result if not d.__contains__('http://dbpedia.org/resource/Category')]
        new_result = []
        for res in result:
            redirect = self.check_redirect(res)
            if redirect:
                new_result.append(redirect[0])
            else:
                new_result.append(res)
        return list(set(new_result))

    def name2entities(self, name):
        s, t0 = self.thing('s')
        t1 = self.triple('?s', self.uri(RDFS.label), self.literal(name))
        t2 = self.triple('?s', self.uri(RDFS.label), self.literal(name.title()))
        t3 = self.redirect(name)
        t4 = self.triple('?s', self.uri(self.demonym), self.literal(name))
        t5 = self.triple('?s', self.uri(self.acronym), self.literal(name))
        t = self.union([t1, t2, t3, t4, t5])
        return self.resource_filter(self.resource_query(s, t + t0))

    def name2entities_expand(self, name):
        s, t0 = self.thing('s')
        t1 = self.triple('?s', self.uri(self.name), self.literal(name))
        t2 = self.triple('?s', self.uri(self.commonName), self.literal(name))
        t3 = self.triple('?s', self.uri(self.longName), self.literal(name))
        t = self.union([t1, t2, t3])
        return self.resource_filter(self.resource_query(s, t+t0))


class QueryGraph(BaseSPARQL):

    """This class implement basic query graph patterns which can query type of things, and a type entity pattern."""

    def __init__(self, result_limit=5000):
        BaseSPARQL.__init__(self, limit=result_limit)

    def type_query(self, concepts, lang='en', show_query=False):
        triples = map(lambda x: self.po_triple(RDF.type, x, 's'), concepts)
        t1 = [t for v, t in triples]
        t1 = self.union(t1)
        v, t2 = self.thing('s')
        return self.text_query(v, t1 + t2, lang, show_query)

    def type_entity_query(self, concepts, entity, show_query=False):
        """
        Construct type entity queries.
        Some more possible patterns.
        ?x ?p1 ?y. ?y ?p2 <%s>.
        ?x ?p1 ?y. <%s> ?p2 ?y.
        ?y ?p1 ?x. ?y ?p2 <%s>
        ?y ?p1 ?x. <%s> ?p2 ?y.
        :param concepts:
        :param entity:
        :return:
        """
        triples = map(lambda x: self.po_triple(RDF.type, x, 's'), concepts)
        t1 = [t for v, t in triples]
        t1 = self.union(t1)
        t2 = self.s_triple(entity, 'p', 's')
        t3 = self.o_triple('s', 'p', entity)
        res1 = self.text_query('s', t1 + t2, show_query=show_query)
        res2 = self.text_query('s', t1 + t3, show_query=show_query)
        return res1 + res2



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2017 Ganggao Zhu- Grupo de Sistemas Inteligentes
# gzhu[at]dit.upm.es
# DIT, UPM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from sematch.utility import FileIO

from nltk.stem import WordNetLemmatizer
import nltk

from collections import Counter
import itertools
import string
import re

stopwords = set(nltk.corpus.stopwords.words('english'))
wn_lemma = WordNetLemmatizer()
reg_tokenizer = nltk.RegexpTokenizer(r'\w+')
word_pattern = re.compile('[a-zA-Z]+')

def word_tokenize(text):
    tokens = reg_tokenizer.tokenize(text)
    return list(filter(lambda token: word_pattern.match(token) and len(token) >= 1, tokens))

def sent_tokenize(text):
    return re.compile(u'[.!?,;:\t\\\\"\\(\\)\\\'\u2019\u2013]|\\s\\-\\s').split(text)

def word_process(words):
    words = [w.lower() for w in words]
    words = [w for w in words if w not in stopwords]
    words = [wn_lemma.lemmatize(w) for w in words]
    words = [wn_lemma.lemmatize(w, pos='v') for w in words]
    return words

chunk_grammar_phrases = """CHUNK: {<NNP><NNP><NNP><NNP>}
    {<NNP|NNPS><NNP|NNPS><NNP|NNPS>}
    {<NNP|NNPS><NNP|NNPS>}
    {<NNP|NNPS>+}"""

class Extraction:

    """This class is used to extract nouns, proper nouns, phrases from text, including descriptions and
    url links"""

    def __init__(self, w_tokenize=None, s_tokenize=None,
                 pos_tag=None, stop_words=None, punct=None,
                 grammar=chunk_grammar_phrases):
        self._word_tokenize = w_tokenize if w_tokenize else word_tokenize
        self._sent_tokenize = s_tokenize if s_tokenize else sent_tokenize
        self._tagger = pos_tag if pos_tag else nltk.pos_tag
        self._stop_words = stop_words if stop_words else stopwords
        self._punct = punct if punct else set(string.punctuation)
        self._chunk_grammar = grammar
        self._chunker = nltk.RegexpParser(self._chunk_grammar)

    def extract_chunks_sent(self, sent):
        """
         Extract chunk phrases from a sentence.
        :param sent: a sentence level text.
        :return: chunk phrases
        """
        tags = self._tagger(self._word_tokenize(sent))
        chunks = nltk.chunk.tree2conlltags(self._chunker.parse(tags))
        # join constituent chunk words into a single chunked phrase
        return [' '.join(word for word, pos, chunk in group)
                  for key, group in itertools.groupby(chunks, lambda (word, pos, chunk): chunk != 'O') if key]

    def extract_chunks_doc(self, text):
        """
        Extract chunk phrases from a document.
        :param text: a document level text
        :return: chunk phrases
        """
        sents = self._sent_tokenize(text)
        sents = [s for s in sents if s]
        return list(itertools.chain.from_iterable(map(self.extract_chunks_sent, sents)))

    def extract_words(self, text, good_tags=set(['NN','NNS','NNP','VBP', 'VBZ','VBG','VBN'])):
        pos_tags = self._tagger(self._word_tokenize(text))
        words = [word for word, tag in pos_tags if tag in good_tags
                 and not all(char in self._punct for char in word)]
        return word_process(words)

    def extract_nouns(self, text):
        return self.extract_words(text, set(['NN', 'NNS', 'NNP']))

    def extract_verbs(self, text):
        return self.extract_words(text, set(['VBP', 'VBZ','VBG','VBN']))

    def category_features(self, categories):
        return map(lambda x:x.replace('http://dbpedia.org/resource/Category:', ''), categories)

    def category2words(self, categories):
        words = []
        for cat in categories:
            cat_str = ' '.join(cat.split('_'))
            words.extend(self.extract_nouns(cat_str))
        return list(set(words))


class RAKE:

    """
    Implementation of RAKE -- Rapid Automatic Keywords Extraction From Individual Documents
    https://github.com/aneesha/RAKE/blob/master/rake.py

    To make a class for rake python implementation for easy incorporation in other modules.
    This class is used to extract keywords from textual document, such as description of entity from DBpedia.
    """

    def __init__(self, stopwords_file='models/FoxStoplist.txt', word_tokenize=None, sent_tokenize=None):
        self._stopwords_pattern = self.build_patterns(self.load_stopwords(stopwords_file))
        self._sent_tokenize = sent_tokenize if sent_tokenize else nltk.sent_tokenize
        self._word_tokenize = word_tokenize if word_tokenize else nltk.word_tokenize

    def load_stopwords(self, filename):
        data = FileIO.read_list_file(FileIO.filename(filename))
        data = [d.split() for d in data[1:]] # skip first line, in case more than one word per line
        data = list(itertools.chain.from_iterable(data))
        return data

    def build_patterns(self, stopwords):
        pattern = lambda x: r'\b' + x + r'(?![\w-])'  # added look ahead for hyphen
        stopword_patterns = map(pattern, stopwords)
        return re.compile('|'.join(stopword_patterns), re.IGNORECASE)

    def candidate_phrases(self, text):
        candidates = []
        for s in self._sent_tokenize(text):
            phrases = re.sub(self._stopwords_pattern, '|', s.strip()).split('|')
            for p in phrases:
                p = p.strip().lower()
                candidates.append(p) if p else None
        return candidates

    def ranking_phrases(self, phrases):
        word_frequency = {}
        word_degree = {}
        for p in phrases:
            words = self._word_tokenize(p)
            degree = len(words) - 1
            for w in words:
                word_frequency.setdefault(w, 0)
                word_frequency[w] += 1
                word_degree.setdefault(w, 0)
                word_degree[w] += degree

        for word in word_frequency:
            word_degree[word] += word_frequency[word]

        word_scorer = lambda x: word_degree[x] / 1.0 * word_frequency[x]
        word_score = {word:word_scorer(word) for word in word_frequency}

        phrase_scorer = lambda x: sum([word_score[word] for word in self._word_tokenize(x)])
        phrase_score = {p:phrase_scorer(p) for p in phrases}

        return phrase_score

    def extract(self, text, ratio=3):
        phrases = self.ranking_phrases(self.candidate_phrases(text))
        phrases = Counter(phrases).most_common(len(phrases.keys()) / ratio)
        phrases, _ = zip(*phrases)
        return phrases





#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2017 Ganggao Zhu- Grupo de Sistemas Inteligentes
# gzhu[at]dit.upm.es
# DIT, UPM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from sematch.semantic.sparql import NameSPARQL, QueryGraph
from sematch.semantic.similarity import YagoTypeSimilarity
from sematch.utility import memoized
from sematch.nlp import word_tokenize, word_process, Extraction

import numpy as np
import itertools
from collections import Counter


class Matcher:

    """This class is used for concept based entity match in DBpedia"""

    def __init__(self, result_limit=5000, expansion=True, show_query=False):
        self._expansion = expansion
        self._show_query = show_query
        self._linker = NameSPARQL()
        self._extracter = Extraction()
        self._yago = YagoTypeSimilarity()
        self._query_graph = QueryGraph(result_limit)

    def type_links(self, word, lang='eng'):
        synsets = self._yago.multilingual2synset(word, lang=lang)
        if self._expansion:
            synsets = list(set(itertools.chain.from_iterable([self._yago.synset_expand(s) for s in synsets])))
        links = []
        for s in synsets:
            link_dic = {}
            link_dic['name'] = s.name()
            link_dic['gloss'] = s._definition
            link_dic['lemma'] = ' '.join(s._lemma_names)
            concept_link = []
            yago_link = self._yago.synset2yago(s)
            dbpedia_link = self._yago.synset2dbpedia(s)
            concept_link.append(yago_link) if yago_link else None
            concept_link.append(dbpedia_link) if dbpedia_link else None
            link_dic['lod'] = concept_link
            if link_dic['lod']:
                links.append(link_dic)
        return links

    def query_process(self, query):
        """
        Process query into concept (common noun) and entity (proper noun). Link them
        to Knowledge Graph uri links respectively.
        :param query: short text query
        :return: tuple of concepts and entities in uris.
        """
        entities = self._extracter.extract_chunks_sent(query)
        entity_filter = list(itertools.chain.from_iterable([e.lower().split() for e in entities]))
        entity_filter = set(entity_filter)
        concepts = list(set(self._extracter.extract_nouns(query)))
        concepts = [c for c in concepts if c not in entity_filter]
        concept_uris = [list(itertools.chain.from_iterable([s['lod'] for s in self.type_links(c)])) for c in concepts]
        concept_uris = list(itertools.chain.from_iterable(concept_uris))
        entity_uris = list(itertools.chain.from_iterable(map(self._linker.name2entities, entities)))
        return list(set(concept_uris)), list(set(entity_uris))

    def match_concepts(self, concepts, lang='en'):
        results = []
        for i in xrange(0, len(concepts), 5):
            results.extend(self._query_graph.type_query(concepts[i:i + 5], lang, self._show_query))
        result_dic = {}
        for res in results:
            if res['uri'] not in result_dic:
                result_dic[res['uri']] = res
        return [result_dic[key] for key in result_dic.keys()]

    def match_type(self, query, lang='eng'):
        lang_map = {'eng':'en','spa':'es', 'cmn':'zh'}
        result_lang = lang_map[lang]
        words = query.split()
        concept_uris = []
        for w in words:
            concepts = list(itertools.chain.from_iterable([s['lod'] for s in self.type_links(w,lang)]))
            concept_uris.extend(concepts)
        concept_uris = list(set(concept_uris))
        return self.match_concepts(concept_uris, result_lang)

    def match_entity_type(self, query):
        results = []
        concepts, entities = self.query_process(query)
        print concepts, entities
        for e in entities:
            for i in xrange(0, len(concepts), 5):
                results.extend(self._query_graph.type_entity_query(concepts[i:i + 5], e, self._show_query))
        result_dic = {}
        for res in results:
            if res['uri'] not in result_dic:
                result_dic[res['uri']] = res
        result = [result_dic[key] for key in result_dic.keys()]
        return result


class SimClassifier:
    """
    This class implements similarity based category classifiers.
    """

    def __init__(self, labels, cat_features, feature_weights, sim_metric, sim_model='weighted'):
        """
         Class initialization.
        :param labels: predefined categories
        :param cat_features: features to represent each category
        :param sim_metric: word similarity function
        """
        self._categories = labels
        self._cat_features = cat_features
        self._feature_weights = feature_weights
        self._sim_metric = sim_metric
        self._sim_model = self.pick_sim_model(sim_model)

    def pick_sim_model(self, sim_model):
        weighted = lambda x, y: self.weighted_similarity(x, y)
        max_sim = lambda x, y: self.max_similarity(x, y)
        average = lambda x, y: self.average_similarity(x, y)
        model_dic = {'weighted':weighted, 'max':max_sim, 'average':average}
        return model_dic[sim_model]

    @classmethod
    def train(cls, corpus, sim_metric, feature_num=5, sim_model='weighted'):
        '''
        Extract categories, features, feature weights, from corpus.
        Compute the weight for each feature token in each category
        The weight is computed as token_count / total_feature_count
        '''
        print "Training..."
        cat_word = {}
        for sent, cat in corpus:
            cat_word.setdefault(cat, []).extend(word_process(word_tokenize(sent)))
        features = {cat: Counter(cat_word[cat]) for cat in cat_word}
        labels = features.keys()
        cat_features = {}
        feature_weights = {}
        for c, f in features.iteritems():
            w_c_pairs = f.most_common(feature_num)
            words, counts = zip(*w_c_pairs)
            cat_features[c] = words
            total_count = float(sum(counts))
            word_weights = []
            for w, count in w_c_pairs:
                word_weights.append((w, count / total_count))
            feature_weights[c] = word_weights
        return cls(labels, cat_features, feature_weights, sim_metric, sim_model)

    def weighted_similarity(self, word, category):
        """
        Input word is compared to each feature word using semantic similarity. The whole similarity
        score is computed as weighted sum.
        :param word: feature word
        :param category: a predefined category
        :return: weighted word similarity score between word and category
        """
        features, weights = zip(*self._feature_weights[category])
        scores = map(lambda x: self._sim_metric(word, x), features)
        return np.dot(np.array(scores), np.array(weights).transpose())

    def max_similarity(self, word, category):
        """
        Compute similarity between word and category, where
        category is represented by several feature words
        :param word: feature word
        :param category: a predefined category
        :return: max word similarity score between word and category
        """
        return max(map(lambda x: self._sim_metric(word, x), self._cat_features[category]) + [0.0])

    def average_similarity(self, word, category):
        """
        Compute similarity between word and category, where
        category is represented by several feature words
        :param word: feature word
        :param category: a predefined category
        :return: average word similarity score between word and category
        """
        sum_score = sum(map(lambda x: self._sim_metric(word, x), self._cat_features[category]) + [0.0])
        N = len(self._cat_features[category])
        return sum_score / N

    @memoized
    def category_similarity(self, word, category):
        """
        Compute the semantic similarity between a word and a category.
        :param word: a feature word
        :param category: predefined category
        :param method: the name of semantic similarity metric
        :return: similarity score between word and category
        """
        return self._sim_model(word, category)

    def classify_single(self, sent, feature_model='max'):
        """
        The input feature words are compared to each category based on category similarity.
        Sum the semantic similarity score between features and category.
        The category having highest similarity score is the correct category.

        :param featuresets: feature sets such as word list
        :param method: specify the semantic similarity metric
        :param model: similarity combination model 'max', 'sum'. Default is 'max'
        :return: the correct category label.
        """
        feature_words = list(set(word_process(word_tokenize(sent))))
        score = {}
        for c in self._categories:
            if feature_model == 'max':
                score[c] = max([self.category_similarity(w, c) for w in feature_words] + [0.0])
            else:
                score[c] = sum([self.category_similarity(w, c) for w in feature_words] + [0.0])
        return Counter(score).most_common(1)[0][0]

    def classify(self, X, feature_model='max'):
        return [self.classify_single(x, feature_model) for x in X]



from sklearn.svm import LinearSVC
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import LabelEncoder
from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.feature_extraction import DictVectorizer

import time

def timeit(func):

    def wrapper(*args, **kwargs):
        start  = time.time()
        result = func(*args, **kwargs)
        delta  = time.time() - start
        return result, delta
    return wrapper


class TextPreprocessor(BaseEstimator, TransformerMixin):
    """
    Transform input text into feature representation
    """
    def __init__(self, corpus, word_sim_metric, feature_num=10, model='sim'):
        """
        :param corpus: use a corpus to train a vector representation
        :param feature_num: number of dimensions
        :param model: onehot or sim
        """
        self._model = model
        self._word_sim = word_sim_metric
        self._features = self.extract_features(corpus, feature_num)

    def fit(self, X, y=None):
        return self

    def inverse_transform(self, X):
        return X

    def extract_features(self, corpus, feature_num=10):
        cat_word = {}
        for sent, cat in corpus:
            cat_word.setdefault(cat, []).extend(word_process(word_tokenize(sent)))
        features = {cat: Counter(cat_word[cat]) for cat in cat_word}
        feature_words = []
        for c, f in features.iteritems():
            words, counts = zip(*f.most_common(feature_num))
            feature_words.extend(list(words))
        feature_words = set(feature_words)
        return feature_words

    def similarity(self, tokens, feature):
        sim = lambda x: self._word_sim(feature, x)
        return max(map(sim, tokens) + [0.0])

    def unigram_features(self, tokens):
        words = set(tokens)
        features = {}
        for f in self._features:
            features['contains({})'.format(f)] = (f in words)
        return features

    def sim_features(self, tokens):
        words = set(tokens)
        features = {}
        for f in self._features:
            features['sim({})'.format(f)] = self.similarity(words, f)
        return features

    def transform(self, X):
        tokenize = lambda x: word_process(word_tokenize(x))
        X_tokens = map(tokenize, X)
        if self._model == 'onehot':
            return map(self.unigram_features, X_tokens)
        else:
            return map(self.sim_features, X_tokens)


class SimSVMClassifier:

    def __init__(self, labels, model):
        self._labels = labels
        self._model = model

    @classmethod
    def train(cls, X, y, word_sim_metric, classifier=LinearSVC,
              feature_num=10, feature_type='sim', verbose=True):

        if isinstance(classifier, type):
            classifier = classifier()

        labels = LabelEncoder()
        y_train = labels.fit_transform(y)

        @timeit
        def build():

            corpus = zip(X, y)
            model = Pipeline([
                ('preprocessor', TextPreprocessor(corpus, word_sim_metric, feature_num, feature_type)),
                ('vectorizer', DictVectorizer()),
                ('classifier', classifier),
            ])

            model.fit(X, y_train)
            return model

        if verbose: print("Building the model")
        model, secs = build()
        if verbose: print("Complete model building in {:0.3f} seconds".format(secs))

        return cls(labels, model)

    def classify(self, X):
        predicted = self._model.predict(X)
        return list(self._labels.inverse_transform(predicted))



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright 2017 Ganggao Zhu- Grupo de Sistemas Inteligentes
# gzhu[at]dit.upm.es
# DIT, UPM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from rdflib import RDF, RDFS, OWL
from sematch.utility import FileIO
import rdflib

class DBpedia:

    """This class implement a wrapper for dbpedia ontology"""

    def __init__(self, src='models/dbpedia_2015-04.owl'):
        self.graph = rdflib.Graph()
        self.graph.parse(FileIO.filename(src))
        self.root = 'http://www.w3.org/2002/07/owl#Thing'
        self.classes = [s for s in self.graph.subjects(RDF.type, OWL.Class)]
        self.o_properties = [s for s in self.graph.subjects(RDF.type, OWL.ObjectProperty)]
        self.d_properties = [s for s in self.graph.subjects(RDF.type, OWL.DatatypeProperty)]
        self.uri2class = {c.toPython():c for c in self.classes}
        self.uri2class[self.root] = rdflib.URIRef(self.root)
        self.class_labels = [self.token(c) for c in self.classes]

    def superClass(self, x):
        return [o.toPython() for s, v, o in self.graph.triples((self.uri2class[x], RDFS.subClassOf, None))]

    def subClass(self, x):
        return [s.toPython() for s, v, o in self.graph.triples((None, RDFS.subClassOf, self.uri2class[x]))]

    def allSubClass(self, x, subList = []):
        for sub in self.subClass(x):
            subList.append(sub)
            self.allSubClass(sub, subList)
        return subList

    def allSuperClass(self, x, superList = []):
        for super in self.superClass():
            superList.append(super)
            self.allSuperClass(super, superList)
        return superList

    def siblingsClass(self, x):
        siblings = []
        for parent in self.superClass(x):
            for child in self.subClass(parent):
                if child != x:
                    siblings.append(child)
        return siblings

    def token(self, x):
        t = [o for o in self.graph.objects(x, RDFS.label) if o.language=='en']
        if t == []:
            return None
        return t[0].toPython()

    def lexicon(self, tag, lst):
        return [(self.token(x), x.n3(), tag) for x in lst if self.token(x)]

    def range(self, p):
        return [o for o in self.graph.objects(p, RDFS.range)]

    def domain(self, p):
        return [o for o in self.graph.objects(p, RDFS.domain)]



from setuptools import setup

setup(name = 'sematch',
      packages=['sematch','sematch.semantic'],
      version = '1.0.2',
      description = 'Semantic similarity framework for knowledge graphs',
      long_description = 'Sematch is an integrated framework for the development, evaluation, and application of semantic similarity for Knowledge Graphs. Semantic similarity is a very useful metric for many applications that require to quantify the semantic similarity between concepts. Sematch focuses on knowledge-based semantic similarity using structural knowledge in taxonomy (e.g. depth, path length, least common subsumer), and statistical information contents (corpus-IC and graph-IC). Many applications share a common pipeline in using semantic similarity analysis. In textual applications, word similarity is measured based on WordNet with various semantic similarity metrics. Then text similarity (sentence level) is measured by composing word similarities, while document level text could be analysed by identifying important sentences based on text similarity, e.g. TextRank. This process starts from measuring concept similarity (word meanings or senses which are denoted as synsets in WordNet) which is normally referred as knowledge-based semantic similarity compared to its conterpart corpus-based approaches that are based on word coocurrence (e.g. Pointwise Mutual Information) or distributional semantics (Explict Semantic Analysis, Word2Vec, GLOVE and etc). Concept similarity are normally quantified using various information such as structural knowledge (e.g. path length, depth, least common subsumer) and statistical knowledge ( information content). KG based applications also meet the requirments in computing concept similarity (e.g. http://dbpedia.org/class/yago/Actor109765278) and entity similarity (e.g. http://dbpedia.org/resource/Madrid). Since entities usually contain several concepts to describe entity meanings, similar to word similarity, entity similarity can be quantified based on their concept similarity. Moreover, those concepts in KGs are usually constructed into hierarchical taxonomies, such as DBpedia ontology class, Wikipedia categories, therefore quantifying concept similarity in WordNet or DBpedia relies on similar semantic information (e.g. path length,  depth, least common subsumer, information content) and semantic similarity metrics (e.g. Path, Wu & Palmer,Li,  Resnik, Lin, Jiang & Conrad and WPath). Consequently,  Sematch provides an integrated framework to develop and evaluate semantic similarity metrics for concepts, words, entities and their applications.',
      author = 'Ganggao Zhu',
      author_email = 'gzhu@dit.upm.es',
      license='Apache 2.0',
      url = 'https://github.com/gsi-upm/sematch',
      keywords = ['semantic similarity', 'taxonomy', 'knowledge graph', 
      'semantic analysis', 'knowledge base', 'WordNet', 'DBpedia','YAGO', 'ontology'],
      include_package_data=True,
      package_data = {'models':['models/dbpedia_2015-04.owl',
                                'models/dbpedia_type_ic.txt',
                                'models/type-linkings.txt',
                                'models/yago_type_ic.txt',
                                'models/FoxStoplist.txt',
                                'models/SmartStoplist.txt'],
                      'dataset/aspect':['dataset/aspect/data.txt'],
                      'dataset/wordsim':['dataset/wordsim/noun_mc.txt',
                                         'dataset/wordsim/noun_rg.txt',
                                         'dataset/wordsim/noun_simlex.txt',
                                         'dataset/wordsim/noun_ws353-sim.txt',
                                         'dataset/wordsim/noun_ws353.txt',
                                         'dataset/wordsim/rg65_EN-ES.txt',
                                         'dataset/wordsim/rg65_spanish.txt',
                                         'dataset/wordsim/graph_mc.txt',
                                         'dataset/wordsim/graph_rg.txt',
                                         'dataset/wordsim/graph_simlex.txt',
                                         'dataset/wordsim/graph_ws353-sim.txt',
                                         'dataset/wordsim/graph_ws353.txt',
                                         'dataset/wordsim/type_mc.txt',
                                         'dataset/wordsim/type_rg.txt',
                                         'dataset/wordsim/type_simlex.txt',
                                         'dataset/wordsim/type_ws353-sim.txt',
                                         'dataset/wordsim/type_ws353.txt',],
                      'dataset/wordsim/results':['dataset/wordsim/results/noun_simlex-wpath.txt'],
                      },
      classifiers = ['Programming Language :: Python :: 2.7',],
      install_requires=['numpy==1.11.0','scipy==0.13.2','scikit-learn==0.17.1','networkx==1.11',
                        'nltk==3.2','rdflib==4.0.1','SPARQLWrapper==1.5.2'],
      )
